/*
 * transport.vala -- taken from libfsoframework
 *
 * Authored by Michael 'Mickey' Lauer <mlauer@vanille-media.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
 */

//===========================================================================
using GLib;
//===========================================================================

//===========================================================================
internal class Transport : Object
//===========================================================================
{
    protected string _name;
    protected uint _speed;
    protected int _fd = -1;

    IOChannel _channel;

    uint _readwatch;
    int _readpriority;
    uint _writeWatch;
    int _writepriority;

    protected HupFunc _hupfunc;
    protected ReadFunc _readfunc;

    protected ByteArray _buffer;

    protected void restartWriter()
    {
        _writeWatch = _channel.add_watch_full( _writepriority, IOCondition.OUT, _writeCallback );
    }

    public void close()
    {
        if ( _readwatch != 0 )
            Source.remove( _readwatch );
        _channel = null;
        if ( _fd != -1 )
            Posix.close( _fd );
        debug( "%s: closed", repr() );
    }

    public virtual string name()
    {
        return _name;
    }

    public void init( string _name, uint _speed, HupFunc? hupfunc, ReadFunc? readfunc, int rp = 0, int wp = 0 )
    {
        _hupfunc = hupfunc;
        _readfunc = readfunc;
        _readpriority = rp;
        _writepriority = wp;
        _buffer = new ByteArray();
    }

    public bool isOpen()
    {
        return ( _fd != -1 );
    }

    public virtual bool open()
    {
        assert( _fd != -1 );
        // setup watches, if we have delegates
        if ( _hupfunc != null || _readfunc != null )
        {
            _channel = new IOChannel.unix_new( _fd );
            _channel.set_encoding( null );
            _channel.set_buffer_size( 32768 );
            _readwatch = _channel.add_watch_full( _readpriority, IOCondition.IN | IOCondition.HUP, _actionCallback );
        }
        // we might have already queued something up in the buffer
        if ( _buffer.len > 0 )
            restartWriter();

        return true;
    }

    public virtual string repr()
    {
        return "<Abstract Transport fd %d>".printf( _fd );
    }

    public int read( void* data, int len )
    {
        assert( _fd != -1 );
        ssize_t bytesread = Posix.read( _fd, data, len );
        return (int)bytesread;
    }

    public int _write( void* data, int len )
    {
        assert( _fd != -1 );
        ssize_t byteswritten = Posix.write( _fd, data, len );
        return (int)byteswritten;
    }

    public int write( void* data, int len )
    {
        var restart = ( _fd != -1 && _buffer.len == 0 );
        var temp = new uint8[len];
        Memory.copy( temp, data, len );
        _buffer.append( temp );

        if ( restart )
            restartWriter();
        return len;
    }

    public void freeze()
    {
        if ( _readwatch != 0 )
        {
            Source.remove( _readwatch );
            _readwatch = 0;
        }
        debug( "frozen" );
    }

    public void thaw()
    {
        if ( _readwatch == 0 )
        {
            _readwatch = _channel.add_watch_full( _readpriority, IOCondition.IN | IOCondition.HUP, _actionCallback );
        }
        debug( "thawn" );
    }

    public bool _actionCallback( IOChannel source, IOCondition condition )
    {
        if ( ( condition & IOCondition.HUP ) == IOCondition.HUP )
        {
            if ( _hupfunc != null )
                _hupfunc( this );
            return false;
        }

        if ( ( condition & IOCondition.IN ) == IOCondition.IN )
        {
            if ( _readfunc != null )
                _readfunc( this );
            return true;
        }

        warning( "actionCallback called with unknown condition: %d", condition );

        return false;
    }

    public bool _writeCallback( IOChannel source, IOCondition condition )
    {
        //debug( "_writeCallback, condition = %d", condition );

        int len = 64 > _buffer.len? (int)_buffer.len : 64;

        var byteswritten = _write( _buffer.data, len  );
        //debug( "_writeCallback: wrote %d bytes", (int)byteswritten );
        _buffer.remove_range( 0, (int)byteswritten );

        return ( _buffer.len != 0 );
    }
}

//===========================================================================
internal class Serial : Transport
//===========================================================================
{
    public Serial( string portname, uint portspeed, HupFunc? hupfunc, ReadFunc? readfunc, int rp = 0, int wp = 0 )
    {
        init( portname, portspeed, hupfunc, readfunc, rp, wp );
        debug( "%s: constructed", repr() );
    }

    public override bool open()
    {
        _fd = Posix.open( _name, Posix.O_RDWR | Posix.O_NOCTTY | Posix.O_NONBLOCK );
        if ( _fd == -1 )
        {
            warning( "%s: could not open %s: %s", repr(), _name, Posix.strerror( Posix.errno ) );
            return false;
        }

        Posix.fcntl( _fd, Posix.F_SETFL, 0 );
        PosixExtra.TermIOs termios = {};
        PosixExtra.tcgetattr( _fd, termios );

        int tspeed;

        switch ( _speed )
        {
            case 0:
                tspeed = PosixExtra.B0;
                break;
            case 50:
                tspeed = PosixExtra.B50;
                break;
            case 75:
                tspeed = PosixExtra.B75;
                break;
            case 110:
                tspeed = PosixExtra.B110;
                break;
            case 134:
                tspeed = PosixExtra.B134;
                break;
            case 150:
                tspeed = PosixExtra.B150;
                break;
            case 200:
                tspeed = PosixExtra.B200;
                break;
            case 300:
                tspeed = PosixExtra.B300;
                break;
            case 600:
                tspeed = PosixExtra.B600;
                break;
            case 1200:
                tspeed = PosixExtra.B1200;
                break;
            case 1800:
                tspeed = PosixExtra.B1800;
                break;
            case 2400:
                tspeed = PosixExtra.B2400;
                break;
            case 4800:
                tspeed = PosixExtra.B4800;
                break;
            case 9600:
                tspeed = PosixExtra.B9600;
                break;
            case 19200:
                tspeed = PosixExtra.B19200;
                break;
            case 38400:
                tspeed = PosixExtra.B38400;
                break;
            case 57600:
                tspeed = PosixExtra.B57600;
                break;
            case 115200:
                tspeed = PosixExtra.B115200;
                break;
            case 230400:
                tspeed = PosixExtra.B230400;
                break;
            case 460800:
                tspeed = PosixExtra.B460800;
                break;
            case 500000:
                tspeed = PosixExtra.B500000;
                break;
            case 576000:
                tspeed = PosixExtra.B576000;
                break;
            case 921600:
                tspeed = PosixExtra.B921600;
                break;
            case 1000000:
                tspeed = PosixExtra.B1000000;
                break;
            case 1152000:
                tspeed = PosixExtra.B1152000;
                break;
            case 1500000:
                tspeed = PosixExtra.B1500000;
                break;
            case 2000000:
                tspeed = PosixExtra.B2000000;
                break;
            case 2500000:
                tspeed = PosixExtra.B2500000;
                break;
            case 3000000:
                tspeed = PosixExtra.B3000000;
                break;
            case 3500000:
                tspeed = PosixExtra.B3500000;
                break;
            case 4000000:
                tspeed = PosixExtra.B4000000;
                break;
            default:
                assert_not_reached();
        }

        PosixExtra.cfsetispeed( termios, tspeed );
        PosixExtra.cfsetospeed( termios, tspeed );

        // local read
        termios.c_cflag |= (PosixExtra.CLOCAL | PosixExtra.CREAD);

        // 8n1
        termios.c_cflag &= ~PosixExtra.PARENB;
        termios.c_cflag &= ~PosixExtra.CSTOPB;
        termios.c_cflag &= ~PosixExtra.CSIZE;
        termios.c_cflag |= PosixExtra.CS8;

        // hardware flow control
        //termios.c_cflag |= PosixExtra.CRTSCTS;

        // software flow control off
        //termios.c_iflag &= ~(PosixExtra.IXON | PosixExtra.IXOFF | PosixExtra.IXANY);

        // raw input
        termios.c_lflag &= ~(PosixExtra.ICANON | PosixExtra.ECHO | PosixExtra.ECHOE | PosixExtra.ISIG);
        termios.c_iflag &= ~(PosixExtra.INLCR | PosixExtra.ICRNL | PosixExtra.IGNCR);

        // raw output
        termios.c_oflag &= ~(PosixExtra.OPOST | PosixExtra.OLCUC | PosixExtra.ONLRET | PosixExtra.ONOCR | PosixExtra.OCRNL );

        /*
        // no special character handling
        termios.c_cc[PosixExtra.VMIN] = 0;
        termios.c_cc[PosixExtra.VTIME] = 2;
        termios.c_cc[PosixExtra.VINTR] = 0;
        termios.c_cc[PosixExtra.VQUIT] = 0;
        termios.c_cc[PosixExtra.VSTART] = 0;
        termios.c_cc[PosixExtra.VSTOP] = 0;
        termios.c_cc[PosixExtra.VSUSP] = 0;
        */
        PosixExtra.tcsetattr( _fd, PosixExtra.TCSANOW, termios);

        /*
        _v24 = PosixExtra.TIOCM_DTR | PosixExtra.TIOCM_RTS;
        Posix.ioctl( _fd, PosixExtra.TIOCMBIS, &_v24 );
        */

        return base.open();
    }

    public override string repr()
    {
        return "<Serial Transport %s @ %u (fd %d)>".printf( _name, _speed, _fd );
    }

}

//===========================================================================
internal class Pty : Transport
//===========================================================================
{

    char[] _ptyname = new char[1024]; // PATH_MAX?

    public Pty( HupFunc? hupfunc, ReadFunc? readfunc, int rp = 0, int wp = 0 )
    {
        init( "Pty", 115200, hupfunc, readfunc, rp, wp );
        debug( "%s: constructed", repr() );
    }

    public override bool open()
    {
        _fd = PosixExtra.posix_openpt( Posix.O_RDWR | Posix.O_NOCTTY | Posix.O_NONBLOCK );
        if ( _fd == -1 )
        {
            warning( "%s: could not open %s: %s", repr(), _name, Posix.strerror( Posix.errno ) );
            return false;
        }

        PosixExtra.grantpt( _fd );
        PosixExtra.unlockpt( _fd );
        PosixExtra.ptsname_r( _fd, _ptyname );

        int flags = Posix.fcntl( _fd, Posix.F_GETFL );
        int res = Posix.fcntl( _fd, Posix.F_SETFL, flags | Posix.O_NONBLOCK );
        if ( res < 0 )
            warning( "%s: can't set pty master to NONBLOCK: %s", repr(), Posix.strerror( Posix.errno ) );

        Posix.fcntl( _fd, Posix.F_SETFL, 0 );

        PosixExtra.TermIOs termios = {};
        PosixExtra.tcgetattr( _fd, termios );

        // local read
        termios.c_cflag |= (PosixExtra.CLOCAL | PosixExtra.CREAD);

        // 8n1
        termios.c_cflag &= ~PosixExtra.PARENB;
        termios.c_cflag &= ~PosixExtra.CSTOPB;
        termios.c_cflag &= ~PosixExtra.CSIZE;
        termios.c_cflag |= PosixExtra.CS8;

        // hardware flow control
        //termios.c_cflag |= PosixExtra.CRTSCTS;

        // software flow control off
        //termios.c_iflag &= ~(PosixExtra.IXON | PosixExtra.IXOFF | PosixExtra.IXANY);

        // raw input
        termios.c_lflag &= ~(PosixExtra.ICANON | PosixExtra.ECHO | PosixExtra.ECHOE | PosixExtra.ISIG);
        termios.c_iflag &= ~(PosixExtra.INLCR | PosixExtra.ICRNL | PosixExtra.IGNCR);

        // raw output
        termios.c_oflag &= ~(PosixExtra.OPOST | PosixExtra.OLCUC | PosixExtra.ONLRET | PosixExtra.ONOCR | PosixExtra.OCRNL );

        /*
        // no special character handling
        termios.c_cc[PosixExtra.VMIN] = 0;
        termios.c_cc[PosixExtra.VTIME] = 2;
        termios.c_cc[PosixExtra.VINTR] = 0;
        termios.c_cc[PosixExtra.VQUIT] = 0;
        termios.c_cc[PosixExtra.VSTART] = 0;
        termios.c_cc[PosixExtra.VSTOP] = 0;
        termios.c_cc[PosixExtra.VSUSP] = 0;
        */
        PosixExtra.tcsetattr( _fd, PosixExtra.TCSANOW, termios);

        /*
        _v24 = PosixExtra.TIOCM_DTR | PosixExtra.TIOCM_RTS;
        Posix.ioctl( _fd, PosixExtra.TIOCMBIS, &_v24 );
        */
        return base.open();
    }

    public override string name()
    {
        return (string)_ptyname;
    }

    public override string repr()
    {
        return "<Pseudo TTY %s (fd %d)>".printf( name(), _fd );
    }
}

//===========================================================================
internal class Socket : Transport
//===========================================================================
{
    uint _port;

    public Socket( HupFunc? hupfunc, ReadFunc? readfunc, int rp = 0, int wp = 0 )
    {
        init( "Pty", 115200, hupfunc, readfunc, rp, wp );
        debug( "%s: constructed", repr() );
    }

    public override string name()
    {
        return "%s:%u".printf( _name, _port );
    }

    public override bool open()
    {
        _fd = Posix.socket( Posix.AF_INET, Posix.SOCK_STREAM, 0 );
        if ( _fd == -1 )
        {
            warning( "%s: could not open %s: %s", repr(), _name, Posix.strerror( Posix.errno ) );
            return false;
        }

        PosixExtra.SockAddrIn sa = {};
        sa.sin_addr.s_addr = PosixExtra.inet_addr( "127.0.0.1" );
        sa.sin_port = PosixExtra.htons( 8080 );
        sa.sin_family = Posix.AF_INET;

        // FIXME listen+accept

        return base.open();
    }

    public override string repr()
    {
        return "<TCP IP %s (fd %d)>".printf( name(), _fd );
    }
}

//===========================================================================
internal delegate void HupFunc( Transport transport );
internal delegate void ReadFunc( Transport transport );
